import logging

import mxnet as mx
from gluoncv.loss import DistillationSoftmaxCrossEntropyLoss
from mxnet import gluon, nd

from autogluon.mxnet.task.metrics import get_metric_instance
from autogluon.mxnet.task.utils import *
from autogluon.core.utils import tqdm
from autogluon.mxnet.utils import collect_params
import autogluon.core as ag
from .processing_params import Sample_params, Getmodel_kwargs
from ..utils.learning_rate import LR_params

__all__ = ['train_image_classification']


@ag.args()
def train_image_classification(args, reporter):
    logging.basicConfig()
    logger = logging.getLogger(__name__)
    if args.verbose:
        logger.setLevel(logging.INFO)
        logger.info(args)

    target_params = Sample_params(args.batch_size, args.num_gpus, args.num_workers)
    batch_size = target_params.get_batchsize
    ctx = target_params.get_context
    classes = args.dataset.num_classes if hasattr(args.dataset, 'num_classes') else None
    target_kwargs = Getmodel_kwargs(ctx,
                                    classes,
                                    args.net,
                                    args.tricks.teacher_name,
                                    args.tricks.hard_weight,
                                    args.hybridize,
                                    args.optimizer.multi_precision,
                                    args.tricks.use_pretrained,
                                    args.tricks.use_gn,
                                    args.tricks.last_gamma,
                                    args.tricks.batch_norm,
                                    args.tricks.use_se)
    distillation = target_kwargs.distillation
    net = target_kwargs.get_net
    input_size = net.input_size if hasattr(net, 'input_size') else args.input_size

    if args.tricks.no_wd:
        for k, v in net.collect_params('.*beta|.*gamma|.*bias').items():
            v.wd_mult = 0.0

    if args.tricks.label_smoothing or args.tricks.mixup:
        sparse_label_loss = False
    else:
        sparse_label_loss = True

    if distillation:
        teacher = target_kwargs.get_teacher

        def teacher_prob(data):
            return [
                nd.softmax(teacher(X.astype(target_kwargs.dtype, copy=False)) / args.tricks.temperature)
                for X in data
            ]

        L = DistillationSoftmaxCrossEntropyLoss(temperature=args.tricks.temperature,
                                                hard_weight=args.tricks.hard_weight,
                                                sparse_label=sparse_label_loss)
    else:
        L = gluon.loss.SoftmaxCrossEntropyLoss(sparse_label=sparse_label_loss)
        teacher_prob = None
    if args.tricks.mixup:
        metric = get_metric_instance('rmse')
    else:
        metric = get_metric_instance(args.metric)

    train_data, val_data, batch_fn, num_batches = get_data_loader(
        args.dataset, input_size, batch_size, args.num_workers, args.final_fit, args.split_ratio
    )

    if isinstance(args.lr_config.lr_mode, str):  # fix
        target_lr = LR_params(args.optimizer.lr, args.lr_config.lr_mode, args.epochs, num_batches,
                              args.lr_config.lr_decay_epoch,
                              args.lr_config.lr_decay,
                              args.lr_config.lr_decay_period,
                              args.lr_config.warmup_epochs,
                              args.lr_config.warmup_lr)
        lr_scheduler = target_lr.get_lr_scheduler
    else:
        lr_scheduler = args.lr_config.lr_mode
    args.optimizer.lr_scheduler = lr_scheduler

    trainer = gluon.Trainer(net.collect_params(), args.optimizer)

    def train(epoch, num_epochs, metric):
        for i, batch in enumerate(train_data):
            metric = default_train_fn(epoch, num_epochs, net, batch, batch_size, L, trainer,
                                      batch_fn, ctx, args.tricks.mixup, args.tricks.label_smoothing,
                                      distillation, args.tricks.mixup_alpha, args.tricks.mixup_off_epoch,
                                      classes, target_kwargs.dtype, metric, teacher_prob)
            mx.nd.waitall()
        return metric

    def test(epoch):
        metric.reset()
        for i, batch in enumerate(val_data):
            default_val_fn(net, batch, batch_fn, metric, ctx, target_kwargs.dtype)
        _, reward = metric.get()
        reporter(epoch=epoch, classification_reward=reward)
        return reward

    # Note: epoch must start with 1, not 0
    tbar = tqdm(range(1, args.epochs + 1))
    for epoch in tbar:
        metric = train(epoch, args.epochs, metric)
        train_metric_name, train_metric_score = metric.get()
        tbar.set_description(f'[Epoch {epoch}] training: {train_metric_name}={train_metric_score :.3f}')
        if not args.final_fit:
            reward = test(epoch)
            tbar.set_description(f'[Epoch {epoch}] Validation: {reward :.3f}')

    if args.final_fit:
        return {'model_params': collect_params(net), 'num_classes': classes}



"""Pascal VOC object detection dataset."""
from __future__ import absolute_import
from __future__ import division
import os
import warnings
import numpy as np
import glob
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
import mxnet as mx
from gluoncv.data.base import VisionDataset

from autogluon.core import *
from .base import DatasetBase

from gluoncv import data as gdata
from gluoncv.utils.metrics.voc_detection import VOC07MApMetric

class CustomVOCDetectionBase(gdata.VOCDetection):
    """Base class for custom Dataset which follows protocol/formatting of the well-known VOC object detection dataset.
    
    Parameters
    ----------
    class: tuple of classes, default = None
        We reuse the neural network weights if the corresponding class appears in the pretrained model. 
        Otherwise, we randomly initialize the neural network weights for new classes.
    root : str, default '~/mxnet/datasets/voc'
        Path to folder storing the dataset.
    splits : list of tuples, default ((2007, 'trainval'), (2012, 'trainval'))
        List of combinations of (year, name)
        For years, candidates can be: 2007, 2012.
        For names, candidates can be: 'train', 'val', 'trainval', 'test'.
    transform : callable, default = None
        A function that takes data and label and transforms them. Refer to
        :doc:`./transforms` for examples.
        A transform function for object detection should take label into consideration,
        because any geometric modification will require label to be modified.
    index_map : dict, default = None
        By default, the 20 classes are mapped into indices from 0 to 19. We can
        customize it by providing a str to int dict specifying how to map class
        names to indices. This is only for advanced users, when you want to swap the orders
        of class labels.
    preload_label : bool, default = True
        If True, then parse and load all labels into memory during
        initialization. It often accelerate speed but require more memory
        usage. Typical preloaded labels took tens of MB. You only need to disable it
        when your dataset is extremely large.
    """

    def __init__(self, classes=None, root=os.path.join('~', '.mxnet', 'datasets', 'voc'),
                 splits=((2007, 'trainval'), (2012, 'trainval')),
                 transform=None, index_map=None, preload_label=True):

        # update classes 
        if classes:
            self._set_class(classes)  
        super(CustomVOCDetectionBase, self).__init__(root=root,
                                               splits=splits,
                                               transform=transform,
                                               index_map=index_map,
                                               preload_label=False),
        self._items_new = [self._items[each_id] for each_id in range(len(self._items)) if self._check_valid(each_id) ]
        self._items = self._items_new
        self._label_cache = self._preload_labels() if preload_label else None
    
    @classmethod
    def _set_class(cls, classes):
        cls.CLASSES = classes
    
    def _load_items(self, splits):
        """Load individual image indices from splits."""
        ids = []
        for subfolder, name in splits:
            root = os.path.join(self._root, subfolder) if subfolder else self._root
            lf = os.path.join(root, 'ImageSets', 'Main', name + '.txt')
            with open(lf, 'r') as f:
                ids += [(root, line.strip()) for line in f.readlines()]
        return ids

    def _check_valid(self, idx):
        """Parse xml file and return labels."""
        img_id = self._items[idx]
        anno_path = self._anno_path.format(*img_id)
        root = ET.parse(anno_path).getroot()
        size = root.find('size')
        width = float(size.find('width').text)
        height = float(size.find('height').text)
        if idx not in self._im_shapes:
            # store the shapes for later usage
            self._im_shapes[idx] = (width, height)
        label = []
        for obj in root.iter('object'):
            try:
                difficult = int(obj.find('difficult').text)
            except ValueError:
                difficult = 0
            cls_name = obj.find('name').text.strip().lower()
            if cls_name not in self.classes:
                continue
            cls_id = self.index_map[cls_name]
            xml_box = obj.find('bndbox')
            xmin = (float(xml_box.find('xmin').text) - 1)
            ymin = (float(xml_box.find('ymin').text) - 1)
            xmax = (float(xml_box.find('xmax').text) - 1)
            ymax = (float(xml_box.find('ymax').text) - 1)

            if not ((0 <= xmin < width) and (0 <= ymin < height) \
                    and (xmin < xmax <= width) and (ymin < ymax <= height)):
                return False

        return True


@obj()
class CustomVOCDetection():
    """Custom Dataset which follows protocol/formatting of the well-known VOC object detection dataset.
    
    Parameters
    ----------
    root : str, default '~/mxnet/datasets/voc'
        Path to folder storing the dataset.
    splits : list of tuples        
        List of combinations of (year, name) to indicate how to split data into training, validation, and test sets.
        For the original VOC dataset, the year candidates can be: 2007, 2012.
        For the original VOC dataset, the name candidates can be: 'train', 'val', 'trainval', 'test'.
        For the original VOC dataset, one might use for example: ((2007, 'trainval'), (2012, 'trainval'))
    classes: tuple of classes
        We reuse the neural network weights if the corresponding class appears in the pretrained model. 
        Otherwise, we randomly initialize the neural network weights for new classes.
    
    Returns
    -------
    Dataset object that can be passed to `task.fit()`, which is actually an :class:`autogluon.space.AutoGluonObject`. 
    To interact with such an object yourself, you must first call `Dataset.init()` to instantiate the object in Python.
    """
    def __init__(self, root, splits, name, classes, **kwargs):
        super().__init__()
        self.root = root
        
        # search classes from gt files for custom dataset
        if not (classes or name):
            classes = self.generate_gt() 
        
        self.dataset = CustomVOCDetectionBase(classes=classes,
                                              root=root,
                                              splits=splits)

        self.metric = VOC07MApMetric(iou_thresh=0.5, class_names=self.dataset.classes)

    def get_dataset_and_metric(self):
        return (self.dataset, self.metric)

    def get_classes(self):
        return self.dataset.classes 

    def generate_gt(self):
        classes = []
        all_xml = glob.glob( os.path.join(self.root, 'Annotations', '*.xml') )
        for each_xml_file in all_xml:
            tree = ET.parse(each_xml_file)
            root = tree.getroot()
            for child in root:
                if child.tag=='object':
                    for item in child:
                        if item.tag=='name':
                            object_name = item.text
                            if object_name not in classes:
                                classes.append(object_name)

        classes = sorted(classes)
        return classes





from gluoncv.utils.parallel import Parallelizable
from mxnet import autograd


class ForwardBackwardTask(Parallelizable):
    def __init__(self, net, optimizer, rpn_cls_loss, rpn_box_loss, rcnn_cls_loss, rcnn_box_loss,
                 mix_ratio, enable_amp):
        super(ForwardBackwardTask, self).__init__()
        self.net = net
        self._optimizer = optimizer
        self.rpn_cls_loss = rpn_cls_loss
        self.rpn_box_loss = rpn_box_loss
        self.rcnn_cls_loss = rcnn_cls_loss
        self.rcnn_box_loss = rcnn_box_loss
        self.mix_ratio = mix_ratio
        self.amp = enable_amp

    def forward_backward(self, x):
        data, label, rpn_cls_targets, rpn_box_targets, rpn_box_masks = x
        with autograd.record():
            gt_label = label[:, :, 4:5]
            gt_box = label[:, :, :4]
            cls_pred, box_pred, roi, samples, matches, rpn_score, rpn_box, anchors, cls_targets, \
                box_targets, box_masks, _ = self.net(data, gt_box, gt_label)
            # losses of rpn
            rpn_score = rpn_score.squeeze(axis=-1)
            num_rpn_pos = (rpn_cls_targets >= 0).sum()
            rpn_loss1 = self.rpn_cls_loss(rpn_score, rpn_cls_targets,
                                          rpn_cls_targets >= 0) * rpn_cls_targets.size / num_rpn_pos
            rpn_loss2 = self.rpn_box_loss(rpn_box, rpn_box_targets,
                                          rpn_box_masks) * rpn_box.size / num_rpn_pos
            # rpn overall loss, use sum rather than average
            rpn_loss = rpn_loss1 + rpn_loss2
            # losses of rcnn
            num_rcnn_pos = (cls_targets >= 0).sum()
            rcnn_loss1 = self.rcnn_cls_loss(cls_pred, cls_targets,
                                            cls_targets.expand_dims(-1) >= 0) * cls_targets.size / \
                num_rcnn_pos
            rcnn_loss2 = self.rcnn_box_loss(box_pred, box_targets, box_masks) * box_pred.size / \
                num_rcnn_pos
            rcnn_loss = rcnn_loss1 + rcnn_loss2
            # overall losses
            total_loss = rpn_loss.sum() * self.mix_ratio + rcnn_loss.sum() * self.mix_ratio

            rpn_loss1_metric = rpn_loss1.mean() * self.mix_ratio
            rpn_loss2_metric = rpn_loss2.mean() * self.mix_ratio
            rcnn_loss1_metric = rcnn_loss1.mean() * self.mix_ratio
            rcnn_loss2_metric = rcnn_loss2.mean() * self.mix_ratio
            rpn_acc_metric = [[rpn_cls_targets, rpn_cls_targets >= 0], [rpn_score]]
            rpn_l1_loss_metric = [[rpn_box_targets, rpn_box_masks], [rpn_box]]
            rcnn_acc_metric = [[cls_targets], [cls_pred]]
            rcnn_l1_loss_metric = [[box_targets, box_masks], [box_pred]]

            if self.amp:
                from mxnet.contrib import amp
                with amp.scale_loss(total_loss, self._optimizer) as scaled_losses:
                    autograd.backward(scaled_losses)
            else:
                total_loss.backward()

        return rpn_loss1_metric, rpn_loss2_metric, rcnn_loss1_metric, rcnn_loss2_metric, \
            rpn_acc_metric, rpn_l1_loss_metric, rcnn_acc_metric, rcnn_l1_loss_metric
